-- Authored 1/27/2008 4:00:16 PM by Gregory Look

----------------------------
--     INITIALIZATION     --
----------------------------


-- Command options
local options = {
	type = "group",
	args = {
		off = {
			name = "offense",
			desc = "Manage the offensive hit table display",
			type = "group",
			args = {
				show = {
					name = "show",
					desc = "Show the offensive hit table display",
					type = "execute",
					func = "ShowOffense",
				},
				hide = {
					name = "hide",
					desc = "Hide the offensive hit table display",
					type = "execute",
					func = "HideOffense",
				},
			},
		},
		def = {
			name = "defense",
			desc = "Manage the defensive hit table display",
			type = "group",
			args = {
				show = {
					name = "show",
					desc = "Show the defensive hit table display",
					type = "execute",
					func = "ShowDefense",
				},
				hide = {
					name = "hide",
					desc = "Hide the defensive hit table display",
					type = "execute",
					func = "HideDefense",
				},
			},
		},
		show = {
			name = "show",
			desc = "Show both hit table displays",
			type = "execute",
			func = function () Hittable:ShowOffense(); Hittable:ShowDefense(); end,
		},
		hide = {
			name = "hide",
			desc = "Hide both hit table displays",
			type = "execute",
			func = function () Hittable:HideOffense(); Hittable:HideDefense(); end,
		},
		vs = {
			name = "versus",
			desc = "Set hit tables' opponent level",
			type = "group",
			args = {
				target = {
					name = "target",
					desc = "Hit tables use your current target for level computation",
					type = "execute",
					func = function () Hittable.vslevel = -1; Hittable:UpdateTablesAndGraphs(); end,
				},
				level = {
					name = "level",
					desc = "Explicitly set opponent level",
					type = "text",
					usage = "level",
					input = true,
					get = false,
					set = function (level) Hittable.vslevel = level; Hittable:UpdateTablesAndGraphs(); end,
					validate = function (level) return (type(level) == "number") and (level > 0); end,
				},
			},
		},
		behind = {
			name = "behind",
			desc = "Assume attacking from behind when not targeted by target",
			type = "toggle",
			get = function () return Hittable.attackfrombehind; end,
			set = function () Hittable.attackfrombehind = not Hittable.attackfrombehind; end,
		},
	},
};

-- Set up addon
Hittable = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0");
Hittable:RegisterChatCommand("/hittable", options);

Hittable.outcome = { };
Hittable.outcome.MISS     = 0;
Hittable.outcome.DODGE    = 1;
Hittable.outcome.PARRY    = 2;
Hittable.outcome.BLOCK    = 3;
Hittable.outcome.GLANCING = 4;
Hittable.outcome.CRITICAL = 5;
Hittable.outcome.CRUSHING = 6;
Hittable.outcome.HIT      = 7;

Hittable.outcome_name = { };
Hittable.outcome_name[Hittable.outcome.MISS    ] = "Miss";
Hittable.outcome_name[Hittable.outcome.DODGE   ] = "Dodge";
Hittable.outcome_name[Hittable.outcome.PARRY   ] = "Parry";
Hittable.outcome_name[Hittable.outcome.BLOCK   ] = "Block";
Hittable.outcome_name[Hittable.outcome.GLANCING] = "Glancing";
Hittable.outcome_name[Hittable.outcome.CRITICAL] = "Critical";
Hittable.outcome_name[Hittable.outcome.CRUSHING] = "Crushing";
Hittable.outcome_name[Hittable.outcome.HIT     ] = "Hit";

Hittable.colors = { };
Hittable.colors[Hittable.outcome.MISS    ] = {0.34, 0.33, 0.33, 1.0};
Hittable.colors[Hittable.outcome.DODGE   ] = {0.00, 0.33, 1.00, 1.0};
Hittable.colors[Hittable.outcome.PARRY   ] = {0.33, 0.00, 1.00, 1.0};
Hittable.colors[Hittable.outcome.BLOCK   ] = {0.00, 1.00, 0.33, 1.0};
Hittable.colors[Hittable.outcome.GLANCING] = {1.00, 1.00, 0.00, 1.0};
Hittable.colors[Hittable.outcome.CRITICAL] = {1.00, 0.00, 0.00, 1.0};
Hittable.colors[Hittable.outcome.CRUSHING] = {1.00, 0.40, 0.00, 1.0};
Hittable.colors[Hittable.outcome.HIT     ] = {1.00, 1.00, 1.00, 1.0};

Hittable.vslevel = -1;
Hittable.attackfrombehind = false;
Hittable.offensive = { };
Hittable.defensive = { };
Hittable.bonusHit = 0;
Hittable.bonusUncrit = 0;
Hittable.canDualWield = nil;

-- OnEnable
function Hittable:OnEnable()
	
	-- register events
	self:RegisterBucketEvent("PLAYER_ENTERING_WORLD",	 0.5, "InitializeData");
	self:RegisterBucketEvent("CHARACTER_POINTS_CHANGED", 0.5, "InitializeData");
	
	self:RegisterBucketEvent("PLAYER_AURAS_CHANGED",	0.5, "UpdateTablesAndGraphs");
	self:RegisterBucketEvent("PLAYER_LEVEL_UP",			0.5, "UpdateTablesAndGraphs");
	self:RegisterBucketEvent("PLAYER_TARGET_CHANGED",	0.5, "UpdateTablesAndGraphs");
	self:RegisterBucketEvent("SKILL_LINES_CHANGED",		0.5, "UpdateTablesAndGraphs");
	self:RegisterBucketEvent("UNIT_AURA",				0.5, "UpdateTablesAndGraphs");
	self:RegisterBucketEvent("UNIT_DEFENSE",			0.5, "UpdateTablesAndGraphs");
	self:RegisterBucketEvent("UNIT_INVENTORY_CHANGED",	0.5, "UpdateTablesAndGraphs");
	
end


--------------------------
--     CALCULATIONS     --
--------------------------

-- normalize the given table
function Hittable:Normalize(t)
	
	if type(t) ~= "table" then return; end
	
	local sum = 0;
	
	for i = 0, getn(t) do
		sum = sum + (t[i] or 0);
	end
	
	for i = 0, getn(t) do
		if t[i] then t[i] = t[i]/sum; end
	end
	
end

-- Calculate added miss based on level differences
function Hittable:GetAdditionalMiss(dskill)
	
	dskill = dskill or 0;
	
	if ( dskill <= 0 ) then
		
		return dskill*0.04;	-- standard defense-based miss
	
	elseif ( dskill <= 10 ) then
		
		return dskill*0.1;	-- gain 0.1% miss per skill difference up to 10 skill higher (2 levels)
		
	else
		
		return dskill*0.4 - 2;	-- jump of 1% miss, plus 0.4% miss per sklil difference
		
	end
	
end
	

-- Calculate average glancing damage penalty versus a given level target
function Hittable:GetGlancingReduction(dskill)
	
	return 8*dskill/5;
	
end

-- get talented hit
function Hittable:GetTalentedHit()
	
	local _, class = UnitClass("player");
	local hit = 0;
	
	if ( class == "PALADIN" ) then
		hit = select(5, GetTalentInfo(2, 3));		-- Precision (Paladin 2, 3) +1% hit per rank
		
	elseif ( class == "ROGUE" ) then
		hit = select(5, GetTalentInfo(2, 6));		-- Precision (Rogue 2, 6) +1% hit per rank
		
	elseif ( class == "SHAMAN" ) then
		hit = 2*select(5, GetTalentInfo(2, 17));	-- Dual Wield Specialization (Shaman 2, 17) +2% hit per rank
		hit = hit + select(5, GetTalentInfo(3, 6));	-- Nature's Guidance (Shaman 3, 6) +1% hit per rank
		
	elseif ( class == "WARRIOR" ) then
		hit = select(5, GetTalentInfo(2, 17));		-- Precision (Warrior 2, 17) +1% hit per rank
		
	end
	
	return hit;
	
end

-- get talented crit reduction
function Hittable:GetTalentedUncrit()
	
	local _, class = UnitClass("player");
	local uncrit = 0;
	
	-- Scorpid Sting (debuff on target) +5% miss
	-- Sleight of Hand (Rogue 3, 3) -2% crit per rank
	
	if ( class == "DRUID" ) then
		uncrit = select(5, GetTalentInfo(2, 16));	-- Survival of the Fittest (Druid 2, 16) -1% crit per rank
		
	elseif ( class == "ROGUE" ) then
		uncrit = 2*select(5, GetTalentInfo(3, 3));	-- Sleight of Hand (Rogue 3, 3) -2% crit per rank
		
	end
	
	return uncrit;
	
end

-- determine whether the player is dualwielding
function Hittable:IsDualWielding()
	
	-- check if the player even has dual wield trained
	if not self.canDualWield and ( GetNumSpellTabs() > 0 ) then
		
		self.canDualWield = false;
		
		local _, _, _, numspells = GetSpellTabInfo(1);
		
		for spell = 1, numspells do
			
			local name = GetSpellName(spell, BOOKTYPE_SPELL);
			
			if ( name == "Dual Wield" ) then self.canDualWield = true; end
			
		end
		
	end
	
	-- if player can't dual wield, then they obviously can't be dual wielding now
	if not self.canDualWield then return false; end
	
	-- if player has zero offhand skill they aren't dual wielding
	-- note an empty offhand due to a 2h equipped gives player's unarmed skill, so we only get 0 here
	-- if they're using a shield or caster offhand etc
	local _, _, ohskill = UnitAttackBothHands("player");
	if ( ohskill < 1 ) then return false; end
	
	-- check if player has a two-hander equipped
	local mainhand = GetInventoryItemLink("player", 16);
	if mainhand and ( select(9, GetItemInfo(mainhand)) == "INVTYPE_2HWEAPON" ) then
		return false;
	end
	
	-- player is probably dual wielding!
	return true;
	
end

-- calculate a basic offensive table
function Hittable:UpdateOffensiveTable(vslevel)
	
	local table = self.offensive;
	
	local level = UnitLevel("player");
	
	if not vslevel or ( vslevel == -1 ) then
		vslevel = level;
		if UnitExists("target") then vslevel = UnitLevel("target"); end
		if vslevel == -1 then vslevel = level + 3; end
	end
	table.vslevel = vslevel;
	
	-- TODO: Intentionally breaking this so I remember, but weapon skill should probably also be capped at your TARGETS' level*5
	
	local mhskill, _, ohskill = UnitAttackBothHands("player");
	mhskill = min(mhskill, 5*level);
	ohskill = min(ohskill, 5*level);
	local skillmod = (mhskill - vslevel*5)*0.04;
	
	-- fill out stats
	table[self.outcome.MISS    ] = self:IsDualWielding() and 24 or 5;	-- add 19% base miss if dual wielding
	table[self.outcome.MISS    ] = table[self.outcome.MISS] + self:GetAdditionalMiss(vslevel*5 - mhskill);
	table[self.outcome.MISS    ] = max(table[self.outcome.MISS] - self.bonusHit - GetCombatRatingBonus(CR_HIT_MELEE), 0);
	table[self.outcome.DODGE   ] = max( 5 - skillmod - GetExpertise()*0.25, 0);
	table[self.outcome.PARRY   ] = max(10 - skillmod - GetExpertise()*0.25, 0);
	table[self.outcome.BLOCK   ] = 0;
	table[self.outcome.GLANCING] = max(8*(vslevel*5 - mhskill)/5, 0);
	table[self.outcome.CRITICAL] = max(GetCritChance() + skillmod, 0);
	table[self.outcome.CRUSHING] = 0;
	table[self.outcome.HIT     ] = 0;
	
	-- if target is not targeting player, assume target is being attacked from behind and remove parry
	if self.attackfrombehind and not UnitIsUnit("player", "targettarget") then
		table[self.outcome.PARRY] = 0;
	end
	
	-- trim table
	local remaining = 100;
	for i = 0, 7 do
		table[i] = min(table[i], remaining);
		remaining = remaining - table[i];
	end
	
	-- expand hit
	table[self.outcome.HIT] = remaining;
	
end

-- calculate a basic defensive table
function Hittable:UpdateDefensiveTable(vslevel)
	
	local table = self.defensive;
	
	local level = UnitLevel("player");
	
	if not vslevel or ( vslevel == -1 ) then
		vslevel = level;
		if UnitExists("target") then vslevel = UnitLevel("target"); end
		if vslevel == -1 then vslevel = level + 3; end
	end
	table.vslevel = vslevel;
	
	local base, bonus = UnitDefense("player");
	local skillmod = (base + bonus - vslevel*5)*0.04;
	
	-- fill out stats
	table[self.outcome.MISS    ] = max(5 + skillmod, 0);
	table[self.outcome.DODGE   ] = GetDodgeChance() - max((vslevel - level)*0.2, 0);
	table[self.outcome.PARRY   ] = GetParryChance() - max((vslevel - level)*0.2, 0);
	table[self.outcome.BLOCK   ] = GetBlockChance() - max((vslevel - level)*0.2, 0);
	table[self.outcome.GLANCING] = 0;
	table[self.outcome.CRITICAL] = max(5 - skillmod - self.bonusUncrit - GetCombatRatingBonus(CR_CRIT_TAKEN_MELEE), 0);
	table[self.outcome.CRUSHING] = max(10*(vslevel - level) - 15, 0);
	table[self.outcome.HIT     ] = 0;
	
	-- if player does not have a weapon in their mainhand, remove parry
	local mainhand = GetInventoryItemLink("player", 16);
	if not mainhand then table[self.outcome.PARRY] = 0; else
		local mainhandtype = select(9, GetItemInfo(mainhand));
		if (mainhandtype ~= "INVTYPE_WEAPON") and
		   (mainhandtype ~= "INVTYPE_WEAPONMAINHAND") and
		   (mainhandtype ~= "INVTYPE_2HWEAPON") then
			table[self.outcome.PARRY] = 0;
		end
	end
	
	-- if player is not wearing a shield, remove block from table
	local offhand = GetInventoryItemLink("player", 17);
	if not offhand then table[self.outcome.BLOCK] = 0; else
		local offhandtype = select(9, GetItemInfo(offhand));
		if offhandtype ~= "INVTYPE_SHIELD" then table[self.outcome.BLOCK] = 0; end
	end
	
	-- Debuffs?
	-- Insect Swarm (debuff on target) +2% miss
	-- Scorpid Sting (debuff on target) +5% miss
	
	-- trim table
	local remaining = 100;
	for i = 0, 7 do
		table[i] = min(table[i], remaining);
		remaining = remaining - table[i];
	end
	
	-- expand hit
	table[self.outcome.HIT] = remaining;
	
end

-- update everything
function Hittable:UpdateTablesAndGraphs()
	
	if self.offensive_frame and self.offensive_frame:IsVisible() then
		self:UpdateOffensiveTable(self.vslevel);
		self:SetTableGraph(self.offensive_frame, self.offensive);
	end
	
	if self.defensive_frame and self.defensive_frame:IsVisible() then
		self:UpdateDefensiveTable(self.vslevel);
		self:SetTableGraph(self.defensive_frame, self.defensive);
	end
	
end

-- update talents/character info
function Hittable:InitializeData()
	
	self.bonusHit = self:GetTalentedHit();
	self.bonusUncrit = self:GetTalentedUncrit();
	self.canDualWield = nil;
	
end

-----------------------
--     INTERFACE     --
-----------------------

-- Show the table graphs
function Hittable:ShowOffense()
	
	if not self.offensive_frame then
		self.offensive_frame = self:CreateTableGraph(UIParent);
		self.offensive_frame.icon:SetTexture("Interface\\Icons\\Ability_Warrior_Challange");
		self:UpdateOffensiveTable(self.vslevel);
		self:SetTableGraph(self.offensive_frame, self.offensive);
		self.offensive_frame:SetPoint("CENTER", parent, "CENTER", 0, 200);
	end
	
	self.offensive_frame:Show();
	
end

function Hittable:ShowDefense()
	
	if not self.defensive_frame then
		self.defensive_frame = self:CreateTableGraph(UIParent);
		self.defensive_frame.icon:SetTexture("Interface\\Icons\\Ability_Warrior_DefensiveStance");
		self:UpdateDefensiveTable(self.vslevel);
		self:SetTableGraph(self.defensive_frame, self.defensive);
		self.defensive_frame:SetPoint("CENTER", parent, "CENTER", 0, -200);
	end
	
	self.defensive_frame:Show();
	
end

function Hittable:HideOffense()
	
	if Hittable.offensive_frame then Hittable.offensive_frame:Hide(); end
	
end

function Hittable:HideDefense()
	
	if Hittable.defensive_frame then Hittable.defensive_frame:Hide(); end
	
end

-- Create a table frame
function Hittable:CreateTableGraph(parent)
	
	local width = 500;
	local height = 64;
	
	-- initialize frame
	local frame = CreateFrame("Frame", nil, parent);
	frame:Hide();
	frame:SetWidth(width+16);
	frame:SetHeight(height+8);
	frame:EnableMouse(true);
	frame:SetMovable(true);
	frame:RegisterForDrag("LeftButton");
	frame:SetScript("OnDragStart", function() this:StartMoving() end);
	frame:SetScript("OnDragStop", function() this:StopMovingOrSizing() end);
	frame:SetBackdrop({
		bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4},
	});
	frame:SetBackdropBorderColor(.5, .5, .5);
	frame:SetBackdropColor(0, 0, 0);
	frame:ClearAllPoints();
	frame:SetPoint("CENTER", parent, "CENTER", 0, 0);
	
	-- create icon
	local icon = frame:CreateTexture(nil, "BORDER");
	icon:SetWidth(height);
	icon:SetHeight(height);
	icon:SetPoint("LEFT", frame, "LEFT", 4, 0);
	--icon:SetBlendMode("ADD");
	icon:SetTexture(1.0, 0.0, 1.0);
	frame.icon = icon;
	
	-- create vs-level text
	local f = GameFontNormal:GetFont();
	local text = frame:CreateFontString(nil, "OVERLAY");
	text:SetFont(f, 14);
	text:SetWidth(30);
	text:SetText("vs\n??");
	text:SetTextColor(1, 1, 0.1);
	text:ClearAllPoints();
	text:SetPoint("BOTTOMRIGHT", icon, "BOTTOMRIGHT", 1, 4);
	frame.vstext = text;
	
	-- create outcome textures
	frame.outcome = { };
	for i = 0, 7 do
		
		local outcome = frame:CreateTexture(nil, "BORDER");
		outcome:SetWidth((width-height)/8);
		outcome:SetHeight(height);
		--outcome:SetBlendMode("ADD");
		outcome:SetTexture(unpack(self.colors[i]));
		if i == 0 then outcome:SetPoint("LEFT", icon, "RIGHT", 0, 0); else
		outcome:SetPoint("LEFT", frame.outcome[i-1], "RIGHT", 0, 0); end
		frame.outcome[i] = outcome;
		
	end
	
	-- create pointers to texts
	frame.pointer = { };
	for i = 0, 7 do
		
		local pointer = frame:CreateTexture(nil, "BORDER");
		pointer:SetWidth(2);
		pointer:SetHeight(height-16);
		pointer:SetTexture(1.0, 1.0, 1.0);
		pointer:SetPoint("CENTER", frame.outcome[i], "CENTER", 0, 0);
		pointer:Hide();
		frame.pointer[i] = pointer;
		
	end	
	
	-- create outcome texts
	frame.text = { };
	for i = 0, 7 do
		
		local text = frame:CreateFontString(nil, "OVERLAY");
		text:SetFont(f, 10);
		text:SetWidth((width-height)/8);
		text:SetText(self.outcome_name[i].."\n12.50%");
		text:SetTextColor(0, 0, 0);
		text:ClearAllPoints();
		text:SetPoint("CENTER", frame.outcome[i], "CENTER", 0, 0);
		frame.text[i] = text;
		
	end
	
	return frame;
	
end

-- Properly set up a table display
function Hittable:SetTableGraph(graph, data)
	
	-- set vs text
	if data.vslevel then
		graph.vstext:SetText("vs\n"..data.vslevel);
	else
		graph.vstext:SetText("vs\n??");
	end
	
	-- normalize the data
	self:Normalize(data);
	
	local height = graph:GetHeight();
	local width = graph:GetWidth() - height;
	local up = 1;
	local anchor = graph.icon;
	
	-- go through each outcome
	for i = 0, getn(data) do
		
		-- Add blocked amount to block text
		if ( i == self.outcome.BLOCK ) then
			graph.text[i]:SetText(string.format("%s\n%.2f%%\n(%d)", self.outcome_name[i], 100*data[i], GetShieldBlock()));
			
		-- Add glancing reduction to glancing text TODO: implement this
		--elseif ( i == self.outcome.GLANCING ) then
			--graph.text[i]:SetText(string.format("%s\n%.2f%%\n(%.1f%%)", self.outcome_name[i], 100*data[i], self:GetGlancingReduction(data.vslevel)));
			
		-- Set standard text format
		else
			graph.text[i]:SetText(string.format("%s\n%.2f%%", self.outcome_name[i], 100*data[i]));
		end
		
		if ( data[i] <= 0 ) then
			
			-- graph.outcome[i]:SetWidth(-1);
			graph.outcome[i]:Hide();
			graph.pointer[i]:Hide();
			graph.text[i]:Hide();
			
		else
			
			graph.outcome[i]:SetWidth(data[i]*width);
			graph.outcome[i]:ClearAllPoints();
			graph.outcome[i]:SetPoint("LEFT", anchor, "RIGHT", 0, 0);
			graph.outcome[i]:Show();
			anchor = graph.outcome[i];
			
			-- check portion width
			if ( data[i] <= 0.08 ) then
				
				graph.pointer[i]:SetPoint("CENTER", graph.outcome[i], "CENTER", 0, height*(1 - 2*up)/2);
				graph.pointer[i]:Show();
				graph.text[i]:SetTextColor(1, 1, 1);
				graph.text[i]:SetPoint("CENTER", graph.outcome[i], "CENTER", 0, height*(1 - 2*up));
				graph.text[i]:Show();
				
				up = 1 - up;
				
			else
				
				graph.pointer[i]:Hide();
				graph.text[i]:SetTextColor(0, 0, 0);
				graph.text[i]:SetPoint("CENTER", graph.outcome[i], "CENTER", 0, 0);
				graph.text[i]:Show();
				
			end
			
		end
		
	end
	
end
